import tensorflow as tf
from tf_neural_net import policy_value_net, N
import numpy
import uff

NUM_RES_BLOCKS = 4
NUM_FILTERS = 32
NUM_INPUT_CHANNELS = 10

def convLayer(input, W_val, layer_num):        
    W = tf.get_variable('cW' + str(layer_num), initializer=W_val)    
    h1 = tf.nn.conv2d(input, W, strides=[1, 1, 1, 1], padding='SAME', data_format='NCHW')
    
    return tf.nn.relu(h1)

def logits(input, W_val, b_val, name):
    W = tf.get_variable(name + '_W', initializer=W_val)
    b = tf.get_variable(name + '_b', initializer=b_val)

    return tf.matmul(input, W_val) + b

def res_block(input, WA_val, WB_val, layer_num):
    WA = tf.get_variable('rWA' + str(layer_num), initializer=WA_val)                        
    h1 = tf.nn.conv2d(input, WA, strides=[1, 1, 1, 1], padding='SAME', data_format='NCHW')

    inter = tf.nn.relu(h1)

    WB = tf.get_variable('rWB' + str(layer_num), initializer=WB_val)
    h1b = tf.nn.conv2d(inter, WB, strides=[1, 1, 1, 1], padding='SAME', data_format='NCHW')    

    return tf.nn.relu(h1b + input)

def fill_conv4_matrix(n_in, n_out, filterSize, GArr):
    s_out = int(n_out/4)
    layer_W = numpy.zeros((filterSize, filterSize, n_in, n_out), dtype=numpy.float32)
    
    layer_W[:,:,:,:s_out] = GArr
    layer_W[:,:,:,s_out:2*s_out] = GArr.swapaxes(0, 1)[:,::-1,:,:]
    layer_W[:,:,:,2*s_out:3*s_out] = GArr[::-1,:,:,:][:,::-1,:,:]
    layer_W[:,:,:,3*s_out:] = GArr.swapaxes(0,1)[::-1,:,:,:]
    
    return layer_W

def fill_4R_matrix(n_in, n_out, filterSize, GArr):
    s_in = int(n_in/4)
    s_out = int(n_out/4)
    
    layer_W = numpy.zeros((filterSize, filterSize, n_in, n_out), dtype=numpy.float32)
    
    layer_W[:,:,:s_in,:s_out] = GArr[0]
    layer_W[:,:,s_in:2*s_in,:s_out] = GArr[3].swapaxes(0, 1)[:,::-1,:,:]
    layer_W[:,:,2*s_in:3*s_in,:s_out] = GArr[2][::-1,:,:,:][:,::-1,:,:]
    layer_W[:,:,3*s_in:,:s_out] = GArr[1].swapaxes(0,1)[::-1,:,:,:]
    
    layer_W[:,:,:s_in,s_out:2*s_out] = GArr[1]
    layer_W[:,:,s_in:2*s_in,s_out:2*s_out] = GArr[0].swapaxes(0, 1)[:,::-1,:,:]
    layer_W[:,:,2*s_in:3*s_in,s_out:2*s_out] = GArr[3][::-1,:,:,:][:,::-1,:,:]
    layer_W[:,:,3*s_in:,s_out:2*s_out] = GArr[2].swapaxes(0,1)[::-1,:,:,:]
    
    layer_W[:,:,:s_in,2*s_out:3*s_out] = GArr[2]
    layer_W[:,:,s_in:2*s_in,2*s_out:3*s_out] = GArr[1].swapaxes(0, 1)[:,::-1,:,:]
    layer_W[:,:,2*s_in:3*s_in,2*s_out:3*s_out] = GArr[0][::-1,:,:,:][:,::-1,:,:]
    layer_W[:,:,3*s_in:,2*s_out:3*s_out] = GArr[3].swapaxes(0,1)[::-1,:,:,:]
    
    layer_W[:,:,:s_in,3*s_out:] = GArr[3]
    layer_W[:,:,s_in:2*s_in,3*s_out:] = GArr[2].swapaxes(0, 1)[:,::-1,:,:]
    layer_W[:,:,2*s_in:3*s_in,3*s_out:] = GArr[1][::-1,:,:,:][:,::-1,:,:]
    layer_W[:,:,3*s_in:,3*s_out:] = GArr[0].swapaxes(0,1)[::-1,:,:,:]
    
    return layer_W

def fill_inv_matrix(n_out, orig_W):
    layer_W = numpy.zeros((N, N, 4, n_out), dtype=numpy.float32)
    n_W = orig_W.reshape((N, N, n_out))

    layer_W[:,:,0,:] = n_W
    layer_W[:,:,1,:] = n_W.swapaxes(0, 1)[:,::-1]
    layer_W[:,:,2,:] = n_W[::-1,:][:,::-1]
    layer_W[:,:,3,:] = n_W.swapaxes(0, 1)[::-1,:]

    return layer_W.reshape((N*N*4, n_out))

if __name__ == "__main__":
    board_input = tf.placeholder(tf.float32, shape=(None, N, N, NUM_INPUT_CHANNELS))
    gen_phase = tf.placeholder(tf.bool)

    with tf.variable_scope("pv_nn"):  
        cur_output = policy_value_net(board_input, N*N, NUM_INPUT_CHANNELS, NUM_FILTERS, gen_phase)                    

    saver = tf.train.Saver(tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, "pv_nn"))
    
    sess = tf.Session()         
    saver.restore(sess, './models/pv_nn_weights_S9_noBV_ResL4_K32-440000')
    
    L1W = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/W0:0"))

    offset = 8
    WA = []
    WB = []
    for ind in range(NUM_RES_BLOCKS):
        WA.append(sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/WA" + str(ind + offset) + ":0")))
        WB.append(sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/WB" + str(ind + offset) + ":0")))

    # value head
    vh_W = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/W12:0"))

    vh_HW = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/vh_W0:0"))
    vh_Hb = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/vh_b0:0"))

    vh_sigL_W = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/sigL_W:0"))
    vh_sigL_b = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/sigL_b:0"))

    # policy head
    ph_W = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/W14:0"))

    ph_pass_W = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/h_W0:0"))
    ph_pass_b = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/h_b0:0"))

    ph_rest_W = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/W15:0"))
    ph_rest_b = sess.run(tf.get_default_graph().get_tensor_by_name("pv_nn/b15:0"))

    sess.close()
    tf.reset_default_graph()


    ##############################################################################################################
    ##############################################################################################################

    # The model is defined after this point, everything above is preprocessing of the weights

    ##############################################################################################################
    ##############################################################################################################


    # make new model
    board_input2 = tf.placeholder(tf.float32, shape=(None, NUM_INPUT_CHANNELS, N, N), name='input_node')
    
    layers = []
    layers.append(convLayer(board_input2, fill_conv4_matrix(NUM_INPUT_CHANNELS, NUM_FILTERS, 3, L1W), layer_num=0))
    
    for i in range(1, NUM_RES_BLOCKS + 1):    
        layers.append(res_block(layers[i-1], fill_4R_matrix(NUM_FILTERS, NUM_FILTERS, 3, WA[i-1]), fill_4R_matrix(NUM_FILTERS, NUM_FILTERS, 3, WB[i-1]), layer_num=i))
    
    # value
    value_head = convLayer(layers[NUM_RES_BLOCKS], fill_4R_matrix(NUM_FILTERS, 4, 3, vh_W), layer_num=1)
        
    value_inv_output_logits = logits(tf.reshape(value_head, [-1, N*N*4]), fill_inv_matrix(64, vh_HW), vh_Hb, 'vh')

    hidden_layer_output = tf.nn.relu(value_inv_output_logits)

    value_output = tf.sigmoid(tf.matmul(hidden_layer_output, vh_sigL_W) + vh_sigL_b, name='v_output')             
        
    # policy
    policy_head = convLayer(layers[NUM_RES_BLOCKS], fill_4R_matrix(NUM_FILTERS, 4, 3, ph_W), layer_num=2)
           
    pass_output = logits(tf.reshape(policy_head, [-1, N*N*4]), fill_inv_matrix(1, ph_pass_W), ph_pass_b, 'ph')
    
    equiv_output_h1 = tf.reduce_mean(tf.nn.conv2d(policy_head, fill_4R_matrix(4, 4, 2*N - 1, ph_rest_W), strides=[1, 1, 1, 1], padding='SAME', data_format='NCHW'), 1)
    equiv_output = tf.reshape(equiv_output_h1 + ph_rest_b, [-1, N*N])

    policy_output_logits = tf.concat([equiv_output, pass_output], 1)
    
    policy_output = tf.nn.softmax(policy_output_logits, name='p_output')              
    
    sess2 = tf.Session()
    sess2.run(tf.global_variables_initializer())

    ## test output
    my_input = numpy.zeros((2, NUM_INPUT_CHANNELS, N, N))
    my_input[0, 3, 3, 0] = 1.0
    my_input[1, 3, 4, 0] = 1.0

    p, v = sess2.run([policy_output, value_output], feed_dict={board_input2 : my_input})    

    print(p)
    print(v)

    for i in tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES):
        print(i)

    ### convert to uff
    output_node_names = ['p_output', 'v_output']
    output_graph_def = tf.graph_util.convert_variables_to_constants(
                sess2, # The session
                tf.get_default_graph().as_graph_def(), 
                output_node_names
    )

    final_output_graph_def = tf.graph_util.remove_training_nodes(output_graph_def)
          
    uff_model = uff.from_tensorflow(final_output_graph_def, 
                                    output_node_names,
                                    intput_node='input_node',
                                    output_filename = "models/full.uff",
                                    text=True)
  
    sess2.close()
    